/*
*   ird: listen to tty connected to IR->RS232 converter
*   
*   94/05/16 john h. dubois iii (john@armory.com)
*   96/07/03 Send help message to fd 1, not 2.
*   
*   The (long) types are for DOS.
*/

#define DEBUG

#ifdef _M_SYSV
#define M_SYSV
#endif

#ifdef M_SYSV
#include <termio.h>
#endif

#include <memory.h>
#include <stdarg.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
/* SCO's DOS cross compiler defines MSDOS */
#ifdef MSDOS
#include <conio.h>
#else
#include <unistd.h>
#endif


#define WriteString(S,fd) write(fd,S,strlen(S))
#define ErrExit(Message) WriteString(Message,2),exit(1)

#ifndef NO_PROTOTYPE
#ifdef MSDOS
int SetPort(char Port[], unsigned long Speed);
#else
int SetTTY(int fd, unsigned long Speed, int Min);
#endif
void DurDisplay(int Bit, unsigned long Count);
void Translate(int Bit, unsigned long Count);
void PeriodDisplay(int Bit, unsigned long Count);
void FindEdges(unsigned long Num, unsigned char *Buf, void (*Action)());
#endif

typedef enum { WAVE, PERIOD, DURATION, TRANSLATED, RAW } Disp;
char *Name;
char *Rep = "_~";
int ZeroMin, ZeroMax, OneMin, OneMax;
int Debug = 0;
extern int errno;
char *Comment = "@(#) ird 1.1 96/07/03";

main(argc,argv)
int argc;
char *argv[];
{
    char *Usage = "Usage: ird [-h] [-c<lh>] [-m<chars>] [-s<speed>] [<tty>]\n";
    unsigned long Speed = 38400;
    int Min = 255;
    Disp DispType = WAVE;
    int InFile = 0;	/* Use fd 0 if none specified */
    int Err = 0;
    unsigned long Jitter = 1;
    char *p;
    char c;			/* Option character */
    extern char *optarg;	/* For optargs */
    extern int optind;		/* For optargs */

    Name = argv[0];
    while ((c = getopt(argc,argv,"hc:s:m:dpj:t:rx")) != -1)
	switch (c) {
	case 'h':
	    WriteStrings(1,
Usage,
"ird reads its input for data sent by an infrared remote control receiver\n"\
"and displays it using characters to show whether carrier was received for\n"\
"each bit time.  Each bit of each input byte is translated into one of two\n"\
"characters.  0 and 1 bits are taken to be non-carrier and periods\n"\
"respectively.  The low bit is expected to contain the first bit time of\n"\
"signal and the high bit the last bit time.\n"\
"If no tty name is given, the standard input is read.\n"\
"Input need not be a real tty (so that log files can be read).\n"\
"The program may be halted by keyboard interrupt.\n"\
"Options:\n"\
"-h: Print this help.\n"\
"-c<lh>: set the characters used to represent a absence and presence of\n"\
"    carrier, e.g.: '-c01'.  If a 0 bit represents the presence of carrier,\n"\
"    give the character for presence first.  The default is '-c_~'.\n"\
"-p: Print the period of each cycle (the period from the onset of each\n"\
"    non-carrier period to the onset of the next non-carrier period).  The\n"\
"    unit is one serial bit time.\n"\
"-d: Print the duration of each carrier and non-carrier period, in\n"\
"    the same units.\n"\
"-r: Output the raw data.\n"\
"-t<0-period,1-period>: Translate data into 0s and 1s.  Cycles with periods\n"\
"    of 0-period and 1-period are translated into 0s and 1s respectively.\n"\
"    The allowable jitter can be set with -j.  Values outside of these\n"\
"    ranges are printed as <period>.  Example: -t 23,46\n"\
"-j<jitter>: Set the allowable jitter.  The default is 1.\n",
#ifndef MSDOS
"-m<chars>: To reduce overhead, batches of <chars> chars are processed.\n"\
"    <chars> must be between 1 and 255.  The default is 255.\n",
#endif
"-s<speed>: Set the tty speed to <speed> Baud.  <speed> must be a valid\n"\
"    tty rate.  The default is 38400.\n", NULL
);
	    exit(0);
	case 's':
	    Speed = (unsigned long) atol(optarg);
	    if (Speed < 50 || Speed > 115200) {
		WriteString("ird: invalid value for speed.\n",2);
		Err = 1;
	    }
	    break;
	case 'j':
	    Jitter = atoi(optarg);
	    if (Jitter < 0 || Jitter > 127) {
		WriteString("ird: invalid jitter value.\n",2);
		Err = 1;
	    }
	    break;
	case 't':
	    ZeroMin = strtol(optarg,&p,10);
	    if (*p)
		OneMin = strtol(p+1,NULL,10);
	    if (ZeroMin <= 0 || OneMin <= 0) {
		WriteString("ird: invalid value given with -t.\n",2);
		Err = 1;
	    }
	    DispType = TRANSLATED;
	    break;
	case 'm':
	    Min = atoi(optarg);
	    if (Min < 1 || Min > 255) {
		WriteString("ird: invalid value given with -m.\n",2);
		Err = 1;
	    }
	    break;
	case 'c':
	    Rep = optarg;
	    if (!Rep[0] || !Rep[1]) {
		WriteString("ird: must give two chars with -c.\n",2);
		Err = 1;
	    }
	case 'r':
	    DispType = RAW;
	    break;
	case 'd':
	    DispType = DURATION;
	    break;
	case 'p':
	    DispType = PERIOD;
	    break;
	case 'x':
	    Debug = 1;
	    break;
	case '?':
	    Err = 1;
	}
    if (Err)
	exit(1);
    if (DispType == TRANSLATED) {
	ZeroMax = ZeroMin + Jitter;
	OneMax = OneMin + Jitter;
	ZeroMin -= Jitter;
	OneMin -= Jitter;
    }
    if (optind < argc && (InFile = open(argv[optind],O_RDONLY)) == -1)
	ErrExit("ird: Could not open input file.\n");
    if (isatty(InFile) && InFile)
#ifdef MSDOS
	(void) SetPort(argv[optind],Speed);
#else
#ifdef M_SYSV
	/* Note: Currently only have termio(M) code! */
	SetTTY(InFile,Speed,Min);
#endif
#endif
    exit(Process(InFile,DispType));
}

void DurDisplay(Bit,Count)
int Bit;
unsigned long Count;
{
    if (!Bit && Count >= 255)	/* End of pulse train */
	putchar('\n');
    else
	printf("%c%ld ",Rep[Bit],Count);
}

void Translate(Bit,Count)
int Bit;
unsigned long Count;
{
    static LowCount = 0;
    static unsigned long Accumulator = 0;

    if (Bit) {
	Count += LowCount;
	if (ZeroMin <= Count && Count <= ZeroMax) {
	    putchar('0');
	    Accumulator <<= 1;
	}
	else if (OneMin <= Count && Count <= OneMax) {
	    putchar('1');
	    Accumulator <<= 1;
	    Accumulator += 1;
	}
	else
	    printf("<%ld>",Count);
    }
    else if (Count >= 255) {	/* End of pulse train */
	printf(" %4lX\n",Accumulator);
	Accumulator = LowCount = 0;
    }
    else
	LowCount = Count;
}

void PeriodDisplay(Bit,Count)
int Bit;
unsigned long Count;
{
    static LowCount = 0;

    if (Bit)
	printf("%ld ",Count + LowCount);
    else if (Count >= 255) {	/* End of pulse train */
	putchar('\n');
	LowCount = 0;
    }
    else
	LowCount = Count;
}


#define BUFSIZE 1024
Process(fd, DispType)
int fd;
Disp DispType;
{
    unsigned char Buffer[BUFSIZE];
    long NRead;

    /* Call processing func once for each read, and once at the end with 0 */
    while ((NRead = read(fd, Buffer, BUFSIZE)) >= 0) {
	switch(DispType) {
	case RAW:
	    write(1,Buffer,NRead);
	    break;
	case TRANSLATED:
	    FindEdges(NRead,Buffer,Translate);
	    break;
	case WAVE:
	    WaveDisplay(NRead,Buffer,Rep);
	    break;
	case DURATION:
	    FindEdges(NRead,Buffer,DurDisplay);
	    break;
	case PERIOD:
	    FindEdges(NRead,Buffer,PeriodDisplay);
	    break;
	}
	if (!NRead)
	    break;
    }
    if (NRead == -1) {
	perror("ird");
	return 1;
    }
    return 0;
}

/*
The converter emits bits in the same order that samples come in.
In RS232, the low bits appear first.  So, the receiving UART makes
the bits that appear first be the low bits.  Thus, we shift bits
out of the low end of the byte and display them in that order.
*/
WaveDisplay(Num,Buf)
unsigned long Num;
unsigned char *Buf;
{
    int Bit;
    unsigned char C;

    for (; Num; Num--) {
	C = *(Buf++);
	for (Bit = 8; Bit; Bit--) {
	    putchar(Rep[C & 1]);
	    C >>= 1;
	}
    }
}

void FindEdges(Num,Buf,Action)
unsigned long Num;
unsigned char *Buf;
void (*Action)();
{
    int Bit;
    int BitsLeft;
    unsigned char C;
    static LastBit = 0;	/* Last bit seen */
    static Count = 0;	/* Number of bits passed since last falling edge */
    unsigned char Skip;

    if (!Num) {	/* Called this way to clean up at EOF (when processing logs) */
	if (Count)
	    (*Action)(LastBit,Count);
	(*Action)(LastBit,0);	/* Tell display routine to clean up */
	Count = LastBit = 0;	/* Re-initialize in case another file */
	return;
    }
    while (Num) {
	/* Search for first char not same as last bit seen */
	Skip = LastBit ? 255 : 0;
	while (Num && *Buf == Skip) {
	    Count += 8;
	    Num--;
	    Buf++;
	}
	if (!Num)	/* No edge found */
	    return;
	C = *(Buf++);
	/* C has at least one edge in it.  Find all edges. */
	for (BitsLeft = 8; BitsLeft--; C >>= 1) {
	    /* Search for edge: first bit not same as last bit seen */
	    if ((C & 1) != LastBit) {	/* Edge found */
		(*Action)(LastBit,Count);
		LastBit = !LastBit;
#ifdef DEBUG
		if (LastBit != (C & 1))
		    ErrExit("Lost sync?!\n");
#endif
		Count = 0;
	    }
	    Count++;
	}
	Num--;	/* Done with that char */
    }
}

WriteStrings(int fd, ...)
{
    va_list ap;
    char *S;

    va_start(ap,fd);
    while (S = va_arg(ap,char *))
	write(fd,S,strlen(S));
    va_end(ap);
}

#ifdef M_SYSV
SetTTY(fd,Speed,Min)
int fd;
unsigned long Speed;
int Min;
{
    struct termio TTYParm;
    int Rates[] = { 0,50,75,110,134,150,200,300,
		    600,1200,1800,2400,4800,9600,19200,38400 };
    int RateInd;

    for (RateInd = 1; RateInd <= 15; RateInd++)
	if (Rates[RateInd] == Speed)
	    break;
    if (RateInd == 16) {
	WriteStrings(2,Name,": Invalid serial rate given.\n",(char *) 0);
	return 1;
    }
    if (ioctl(fd,TCGETA,&TTYParm)) {
	WriteStrings(2,Name,": Could not get tty parameters: ",
	strerror(errno),"\n",(char *) 0);
	return 1;
    }
    TTYParm.c_cflag = TTYParm.c_cflag & ~CBAUD | RateInd;
    TTYParm.c_lflag &= ~ICANON;
    TTYParm.c_cc[VMIN] = Min;
    TTYParm.c_cc[VTIME] = 1;	/* wait 1/10 second for more data */
    if (ioctl(fd,TCSETA,&TTYParm)) {
	WriteStrings(2,Name,": Could not set tty parameters: ",
	strerror(errno),"\n",(char *) 0);
	return 1;
    }
    if (Debug)
	fprintf(stderr,"Set tty to %d Baud, vmin=%d, vtime=1\n",Speed,Min);
    return 0;
}
#endif
#ifdef MSDOS
/*
    This code is needed because the DOS cross-compiling libraries
    and the BIOS don't define speeds above 9600.
*/
SetPort(Port,Speed)
char Port[];
unsigned long Speed;
{
    unsigned BaseAddr;
    unsigned long Divisor;
    unsigned BaseAddrs[] = { 0x3f8, 0x2f8, 0x3e8, 0x2e8 };

    /* Port name must be com1..4, with or without a trailing ':' */
    if (strnicmp(Port,"com",3) || !(Port[3] >= '1') || !(Port[3] <= '4')
    || !(!Port[4] || Port[4] == ':')) {
	WriteString(
	"ird: Warning: Unknown device name; could not set port parameters.\n",
	2);
	return(1);
    }
    BaseAddr = BaseAddrs[Port[3] - '1'];
    Divisor = 115200 / Speed;
    /* Fail if not an integer fraction of the base sio rate
    (this program will only be used at high speeds that will divide evenly) */
    if (115200 % Speed) {
	WriteString("ird: Warning: invalid speed; port parameters not set.\n",
	2);
	return(1);
    }
    if (Debug)
	printf(
	"Setting com%d (base addr 0x%X) to speed %lu with divisor %lu.\n",
	Port[3] - '0',BaseAddr,Speed,Divisor);
    outpw(BaseAddr + 1,(unsigned) Divisor);
}
#endif
